@@ -1,3 +1,11 @@
+from validio_sdk.exception import (
+    ConfigInvalidError,
+    ConfigNotFoundError,
+    UnauthorizedError,
+    ValidioConnectionError,
+    ValidioError,
+)
+
 # Since we don't want to refer to or use types defined in the `resource` module
 # outside of IaC we re-export these types so we can refer to them from the SDK
 # with a more logical import path.
@@ -8,6 +16,7 @@ from validio_sdk.resource.filters import (
     EnumFilterOperator,
     NullFilter,
     NullFilterOperator,
+    SqlFilter,
     StringFilter,
     StringFilterOperator,
     ThresholdFilter,
@@ -18,13 +27,19 @@ from validio_sdk.util import load_jtd_schema
 __all__ = [
     "BooleanFilter",
     "BooleanFilterOperator",
+    "ConfigInvalidError",
+    "ConfigNotFoundError",
     "EnumFilter",
     "EnumFilterOperator",
     "NullFilter",
     "NullFilterOperator",
+    "SqlFilter",
     "StringFilter",
     "StringFilterOperator",
     "ThresholdFilter",
     "ThresholdFilterOperator",
+    "UnauthorizedError",
+    "ValidioConnectionError",
+    "ValidioError",
     "load_jtd_schema",
 ]
