@@ -1,4 +1,7 @@
-class ManifestConfigurationError(Exception):
+from validio_sdk.exception import ValidioError
+
+
+class ManifestConfigurationError(ValidioError):
     """Raised when there is an invalid configuration in the manifest."""
 
 
@@ -14,21 +17,6 @@ def updated_resource_type_mismatch_exception(
     )
 
 
-def rejected_update_exception(
-    resource_name: str, field_name: str, manifest_value: object, server_value: object
-) -> ManifestConfigurationError:
-    """
-    Raised when an attempt is made to update an immutable
-    value (e.g. changing source_id on a validator).
-    """
-    return ManifestConfigurationError(
-        f"invalid configuration for resource '{resource_name}': "
-        f"field '{field_name}' is immutable; "
-        f"manifest value is '{manifest_value}'; "
-        f"server value is '{server_value}'"
-    )
-
-
 def max_resource_depth_exceeded(
     resource_name: str,
 ) -> ManifestConfigurationError:
