@@ -1,15 +1,23 @@
 import dataclasses
 import inspect
+import re
 from abc import abstractmethod
 from dataclasses import dataclass
-from typing import TYPE_CHECKING, Any, Optional, TypeVar, cast
+from functools import wraps
+from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar, cast
 
-from camel_converter import to_snake
+from camel_converter import to_camel, to_snake
 
 # We need validio_sdk in scope due to eval.
 # ruff: noqa: F401
 import validio_sdk
-from validio_sdk.resource._diffable import Diffable
+from validio_sdk._api.api import APIClient, execute_mutation
+from validio_sdk.client import Session
+from validio_sdk.exception import ValidioError
+from validio_sdk.resource._diffable import (
+    ApiSecretChangeNestedResource,
+    Diffable,
+)
 from validio_sdk.resource._serde import (
     NODE_TYPE_FIELD_NAME,
     ImportValue,
@@ -19,23 +27,23 @@ from validio_sdk.resource._serde import (
     _import_value_repr,
     without_skipped_internal_fields,
 )
-from validio_sdk.validio_client import ValidioAPIClient
 
 if TYPE_CHECKING:
     from validio_sdk.code._import import ImportContext
     from validio_sdk.resource.channels import Channel
     from validio_sdk.resource.credentials import Credential
-    from validio_sdk.resource.destinations import Destination
-    from validio_sdk.resource.identity_providers import IdentityProvider
+    from validio_sdk.resource.filters import Filter
     from validio_sdk.resource.notification_rules import NotificationRule
     from validio_sdk.resource.segmentations import Segmentation
     from validio_sdk.resource.sources import Source
-    from validio_sdk.resource.users import User
+    from validio_sdk.resource.tags import Tag
     from validio_sdk.resource.validators import Validator
     from validio_sdk.resource.windows import Window
 
 R = TypeVar("R", bound="Resource")
 
+CREATE_ONLY_RESOURCES = {"tags"}
+
 
 @dataclass
 class DiffContext:
@@ -44,20 +52,17 @@ class DiffContext:
     revisit them at a later point. e.g. to compare two graphs.
     """
 
+    tags: dict[str, "Tag"] = dataclasses.field(default_factory=dict)
     credentials: dict[str, "Credential"] = dataclasses.field(default_factory=dict)
     channels: dict[str, "Channel"] = dataclasses.field(default_factory=dict)
     sources: dict[str, "Source"] = dataclasses.field(default_factory=dict)
-    destinations: dict[str, "Destination"] = dataclasses.field(default_factory=dict)
     windows: dict[str, "Window"] = dataclasses.field(default_factory=dict)
+    filters: dict[str, "Filter"] = dataclasses.field(default_factory=dict)
     segmentations: dict[str, "Segmentation"] = dataclasses.field(default_factory=dict)
     validators: dict[str, "Validator"] = dataclasses.field(default_factory=dict)
     notification_rules: dict[str, "NotificationRule"] = dataclasses.field(
         default_factory=dict
     )
-    identity_providers: dict[str, "IdentityProvider"] = dataclasses.field(
-        default_factory=dict
-    )
-    users: dict[str, "User"] = dataclasses.field(default_factory=dict)
 
     # Validators objects that are yet to be decoded
     pending_validators_raw: dict[str, tuple[type, dict[str, Any]]] = dataclasses.field(
@@ -73,27 +78,39 @@ class DiffContext:
         ]
 
     @staticmethod
-    def fields_topological_order() -> list[tuple[str, list[str]]]:
-        """Returns the fields but with their parent-child dependency encoded."""
+    def fields_singular() -> set[str]:
+        return {f[: len(f) - 1] for f in DiffContext.fields()}
+
+    @staticmethod
+    def fields_topological_order(
+        graph_mode: bool = False,
+    ) -> list[tuple[str, list[str]]]:
+        """
+        Returns the fields but with their parent-child dependency encoded.
+        If graph_mode is turned off (default), then the returned order is
+        a tree datastructure (each node has a single parent) and will be
+        visited exactly once.
+        """
         fields: list[tuple[str, list[str]]] = [
-            ("identity_providers", []),
-            ("users", []),
             # Channels need to come before sources because notification rules
             # have references to sources.
+            ("tags", ["sources", "validators"]),
             ("channels", ["notification_rules"]),
-            ("credentials", ["sources", "destinations"]),
-            ("sources", ["windows", "segmentations", "validators"]),
-            ("destinations", []),
-            ("windows", []),
-            ("segmentations", []),
+            ("credentials", ["sources"]),
+            ("sources", ["windows", "segmentations", "filters", "validators"]),
+            ("windows", [] if not graph_mode else ["validators"]),
+            ("segmentations", [] if not graph_mode else ["validators"]),
+            ("filters", [] if not graph_mode else ["segmentations", "validators"]),
             ("validators", []),
             ("notification_rules", []),
         ]
         # Sanity check that this doesn't go out of sync
         parents = {parent for (parent, _) in fields}
         expected = set(DiffContext.fields())
+
         if parents != expected:
-            raise Exception(f"DiffContext fields mismatch {parents} != {expected}")
+            raise ValidioError(f"DiffContext fields mismatch {parents} != {expected}")
+
         return fields
 
 
@@ -123,6 +140,25 @@ class ResourceID:
         return None
 
 
+@dataclass(order=True)
+class ResourceDeprecation:
+    """Deprecation information for a resource."""
+
+    resource_type: str
+    resource_name: str
+    message: str
+
+    def __str__(self) -> str:
+        """String implementation."""
+        named = f" named '{self.resource_name}'" if self.resource_name else ""
+
+        return f"On '{self.resource_type}'{named}: {self.message}"
+
+    def _encode(self) -> dict[str, str]:
+        """Encoder."""
+        return self.__dict__
+
+
 class Resource(Diffable):
     """
     Dataclass representing a resource object.
@@ -131,12 +167,19 @@ class Resource(Diffable):
     between resources.
     """
 
-    def __init__(self, name: str, __internal__) -> None:
+    def __init__(
+        self,
+        name: str,
+        display_name: str | None,
+        __internal__: "ResourceGraph",
+        ignore_changes: bool,
+    ):
         self._id: ResourceID = ResourceID()
         self._namespace: str | None = None
         self._node_type: str = self.__class__.__name__
 
         self.name = name
+        self.display_name = display_name if display_name else name
 
         # Split by resource type. e.g a Source will have different children sets
         # per resource type (window, segmentation, validators)
@@ -159,9 +202,85 @@ class Resource(Diffable):
         # on the server.
         self._applied = False
 
+        # If set to true, changes to the resource will be ignored
+        self.ignore_changes = ignore_changes
+
+        # TODO(VR-3875): Fully remove after deprecation warning period.
+        if hasattr(self, "_arg_deprecation"):
+            self.add_deprecation(getattr(self, "_arg_deprecation"))
+            delattr(self, "_arg_deprecation")
+
+    # TODO(VR-3875): Fully remove after deprecation warning period.
+    def _check_and_delete_arg_deprecation_on_child(
+        self, child: "Diffable | Resource | None"
+    ) -> None:
+        """Add child object deprecation.
+
+        Some child objects such as thresholds or notification rule conditions
+        aren't resources and thus can't add deprecations to themselves. Instead
+        when we have deprecation we usually add it to the parent resource where
+        used.
+
+        This method on a parent class accepts a child obejct that can contain an
+        internal field named `_arg_deprecation` and if it does, a deprecation
+        notice will be added referring to the parent class.
+        """
+        if hasattr(child, "_arg_deprecation"):
+            self.add_deprecation(
+                f"(Used on '{self.__class__.__name__}' named '{self.name}'): "
+                + getattr(child, "_arg_deprecation"),
+                child,
+            )
+            delattr(child, "_arg_deprecation")
+
+    def _mutable_fields(self) -> set[str]:
+        return {"display_name"}
+
+    def add_field_deprecation(
+        self, current_field: str, suggested_field: str | None = None
+    ) -> None:
+        """Mark a field as deprecated.
+
+        :param current_field: The field to deprecated
+        :param suggested_field: Potentially new field replacing it
+        """
+        description = f"Field '{current_field}' is deprecated"
+        if suggested_field:
+            description += f", please use '{suggested_field}' instead"
+
+        self.add_deprecation(description)
+
+    def add_deprecation(
+        self,
+        description: str,
+        diff_object: "Diffable | Resource | None" = None,
+    ) -> None:
+        """Add a deprecation warning for a resource.
+
+        :param description: Deprecation description
+        :param diff_object: If the deprecation is on a `Diffable` used by this
+            parent we can pass that to print it to the right resource.
+        """
+        if diff_object is None:
+            resource_type = self.__class__.__name__
+            resource_name = self.name
+        else:
+            resource_type = diff_object.__class__.__name__
+            resource_name = (
+                diff_object.name if isinstance(diff_object, Resource) else ""
+            )
+
+        self._resource_graph.add_deprecation(
+            ResourceDeprecation(
+                resource_name=resource_name,
+                resource_type=resource_type,
+                message=description,
+            )
+        )
+
     def _must_id(self) -> str:
         if self._id.value is None:
-            raise RuntimeError(
+            raise ValidioError(
                 f"resource {self.__class__.__name__}(name={self.name}) "
                 "has unresolved ID"
             )
@@ -170,7 +289,7 @@ class Resource(Diffable):
 
     def _must_namespace(self) -> str:
         if self._namespace is None:
-            raise RuntimeError(
+            raise ValidioError(
                 f"resource {self.__class__.__name__}(name={self.name}) "
                 "has unresolved namespace"
             )
@@ -178,7 +297,7 @@ class Resource(Diffable):
         return self._namespace
 
     # Adds the specified resource as a 'child' of the self resource.
-    def add(self, resource_name: str, child: "Resource"):
+    def add(self, resource_name: str, child: "Resource") -> None:
         self._resource_graph.allocate_resource_name(resource_name, child)
 
         child_type: str = child.resource_class_name()
@@ -208,97 +327,107 @@ class Resource(Diffable):
 
         return {"_children": children}
 
-    def _nested_objects(self) -> dict[str, Optional["Diffable"]]:
+    def _nested_objects(self) -> dict[str, Optional["Diffable | list[Diffable]"]]:
+        return {}
+
+    def _nested_mutable_parents(self) -> dict[str, str | None]:
+        """
+        Returns any nested mutable parent resources.
+        This has information about any 'nested' fields that
+        point to parent resource references. (e.g a validator
+        pointing to a reference filter).
+
+        Returned map contains field_name => field_value
+        """
         return {}
 
     def _api_create_response_field_name(self) -> str:
         return to_snake(self.resource_class_name())
 
-    def _api_delete_method_name(self) -> str:
-        return to_snake(self.resource_class_name())
+    def _api_create_method_name(self) -> str:
+        name = self.__class__.__name__
+        lc_first = name[0].lower() + name[1:]
+
+        return f"{lc_first}Create"
+
+    def _api_update_method_name(self) -> str:
+        name = self.__class__.__name__
+        lc_first = name[0].lower() + name[1:]
+
+        return f"{lc_first}Update"
+
+    def _api_create_arguments(self) -> dict[str, str]:
+        return {
+            "input": f"{self.__class__.__name__}CreateInput!",
+        }
+
+    def _api_update_arguments(self) -> dict[str, str]:
+        return {"input": f"{self.__class__.__name__}UpdateInput!"}
+
+    def _api_delete_arguments(self) -> dict[str, str]:
+        return {"ids": f"[{self.resource_class_name()}Id!]!"}
 
     async def _api_create(
-        self, namespace: str, client: ValidioAPIClient, ctx: "DiffContext"
+        self,
+        namespace: str,
+        ctx: "DiffContext",
+        session: Session,
     ) -> str:
         """
         Create the resource, and resolve's the current instance with
         the ID assigned that was assigned by the server.
         """
+        if self.has_user_defined_name() and not re.match(
+            r"^[a-z0-9_.-]{2,251}$", self.name, re.IGNORECASE
+        ):
+            raise ValidioError(
+                f"invalid resource name '{self.name}': must be 2-253 "
+                "characters containing only a-z, A-Z, 0-9, _, - or ."
+            )
+
         create_input = self._api_create_input(namespace, ctx)
-        payload = await self._api_create_or_update("create", client, create_input)
-        id_ = payload.id
+        payload = await self._api_create_or_update(
+            "create", create_input, session=session
+        )
+
+        id_ = payload["id"]
         self._id.value = id_
+
         return id_
 
     async def _api_update(
-        self, namespace: str, client: ValidioAPIClient, ctx: "DiffContext"
-    ):
+        self,
+        namespace: str,
+        ctx: "DiffContext",
+        session: Session,
+    ) -> None:
         """Perform api call to update the resource."""
-        from validio_sdk.resource.validators import Validator
-
         update_input = self._api_update_input(namespace, ctx)
-        await self._api_create_or_update("update", client, update_input)
-
-        if not isinstance(self, Validator):
-            return
-
-        # For validators, update the threshold as well. Those have
-        # an explicit update api.
-        resource_snake_case = to_snake(self.threshold.__class__.__name__)
-        method_name = f"update_validator_with_{resource_snake_case}"
-        method_fn = client.__getattribute__(method_name)
-
-        response = await method_fn(self.threshold._api_update_input(self._must_id()))
-        self._check_graphql_response(
-            response=response,
-            method_name=method_name,
-            response_field=None,
-        )
-
-    async def _api_delete(self, client: ValidioAPIClient) -> Any:
-        """Api method to delete this resource."""
-        method_name = f"delete_{self._api_delete_method_name()}"
-        method_fn = client.__getattribute__(method_name)
-        response = await method_fn(self._id.value)
-        return self._check_graphql_response(
-            response=response,
-            method_name=method_name,
-            response_field=None,
-        )
+        await self._api_create_or_update("update", update_input, session=session)
 
     async def _api_create_or_update(
-        self, verb: str, client: ValidioAPIClient, api_input: Any | None
-    ):
-        from validio_sdk.resource.validators import Validator
+        self,
+        verb: str,
+        api_input: Any | None,
+        session: Session,
+    ) -> Any:
+        if verb == "create":
+            method_name = self._api_create_method_name()
+            arguments = self._api_create_arguments()
+        else:
+            method_name = self._api_update_method_name()
+            arguments = self._api_update_arguments()
 
-        """
-        Perform api call to create or update the resource.
-        """
-        # Resource names map, for the most part, directly to the equivalent
-        # graphql operation. So we strive to use the same logic for all resource
-        # types, by templating the operation name and parameters.
-
-        # e.g GcpBigQuerySource
-        class_name = self.__class__.__name__
-        # => gcp_big_query_source
-        resource_snake_case = to_snake(class_name)
-
-        # Validators end with `_with_dynamic_threshold_` on their create apis.
-        method_suffix = (
-            f"_with_{to_snake(self.threshold.__class__.__name__)}"
-            if isinstance(self, Validator) and verb == "create"
-            else ""
-        )
-        # => create_gcp_big_query_source
-        method_name = f"{verb}_{resource_snake_case}{method_suffix}"
-        method_fn = client.__getattribute__(method_name)
-        # => source
         response_field = self._api_create_response_field_name()
 
-        if type(api_input) is dict:
-            response = await method_fn(**api_input)
-        else:
-            response = await method_fn(api_input)
+        response = await execute_mutation(
+            session,
+            method_name,
+            arguments,
+            api_input,
+            response_field,
+            returns="id",
+        )
 
         return self._check_graphql_response(
             response=response,
@@ -311,39 +440,46 @@ class Resource(Diffable):
         response: Any,
         method_name: str,
         response_field: str | None,
-    ) -> Any | None:
-        if len(response.errors) > 0:
-            raise RuntimeError(
+    ) -> Any:
+        errors = response.get("errors")
+        if errors:
+            raise ValidioError(
                 f"operation '{method_name}' failed for "
                 f"resource {self.__class__.__name__}(name={self.name}): "
-                f"{response.errors}"
+                f"{errors}"
             )
 
         if response_field is None:
             return None
 
-        payload = response.__getattribute__(response_field)
-        if payload is None:
-            raise RuntimeError(
+        if response_field not in response:
+            raise Exception(f"Unexpected response: {response}")
+
+        if not response.get(response_field):
+            raise ValidioError(
                 f"operation '{method_name}' failed for '"
                 f"resource {self.__class__.__name__}(name={self.name}): '"
                 "missing response body"
             )
 
-        return payload
+        return response[response_field]
 
-    def _api_create_input(self, namespace: str, _: "DiffContext") -> Any:
+    def _api_create_input(self, _namespace: str, _: "DiffContext") -> Any:
         """
         Returns the graphql input(s) to create this resource
         Returned value can either be the <Resource>CreateInput instance or
         if the create api takes in several arguments, then a dict[str, <Input>]
         that will be passed to the api.
 
-        Default behavior (which should be overriden in most resource types) takes
+        Default behavior (which should be overridden in most resource types) takes
         all fields on the resource, assumes that the field names match 1-1 with the
         corresponding input.
+
+        Only root level objects such as credentials and channels should return
+        the namespace, child resources will always inherit this value from its
+        parent.
         """
-        return _api_create_input_params(self, namespace)
+        return _api_create_input_params(self)
 
     def _api_update_input(self, _namespace: str, _: "DiffContext") -> Any:
         """
@@ -352,15 +488,34 @@ class Resource(Diffable):
         """
         return _api_update_input_params(self)
 
+    def _api_delete_input(self) -> Any:
+        """
+        Returns the graphql input(s) to delete this resource. Most APIs do bulk
+        deletes so we pass our own ID as a list but not all resources work this
+        way.
+        """
+        return {"ids": [self._must_id()]}
+
     def _import_str(
         self,
         indent_level: int,
         import_ctx: "ImportContext",
         inits: list[tuple[str, Any, str | None]] | None = None,
+        skip: set[str] | None = None,
     ) -> str:
-        params = [("name", repr(self.name), None), *list(inits or [])]
+        skip = set() if skip is None else skip
+        params: list[tuple[str, Any, str | None]] = []
+
+        if self.has_user_defined_name():
+            params = [("name", repr(self.name), None)]
+
+        if inits:
+            params.extend(inits)
 
         for field, import_value in self._import_params().items():
+            if field in skip:
+                continue
+
             params.append(
                 (
                     field,
@@ -376,6 +531,9 @@ class Resource(Diffable):
     def _import_params(self) -> dict[str, ImportValue]:
         return _import_resource_params(resource=self)
 
+    def has_user_defined_name(self) -> bool:
+        return True
+
     @abstractmethod
     def resource_class_name(self) -> str:
         """What type of resource this is. (e.g. Window, Segmentation etc)."""
@@ -384,6 +542,37 @@ class Resource(Diffable):
     def _encode(self) -> dict[str, object]:
         """Encode the resource as json."""
 
+    def _api_secret_change_input(self) -> dict[str, Any]:
+        nested_objects = {
+            to_camel(f): obj._api_input()
+            for f, obj in self._nested_secret_objects().items()
+        }
+
+        return {
+            "input": {
+                "id": self._must_id(),
+                **{to_camel(f): getattr(self, f) for f in self._secret_fields()},
+                **nested_objects,
+            }
+        }
+
+    def _api_secret_change_response(
+        self,
+    ) -> dict[str, None | ApiSecretChangeNestedResource]:
+        return {
+            **{to_camel(secret_field): None for secret_field in self._secret_fields()},
+            **self._nested_secret_objects(),
+        }
+
+    def _nested_secret_objects(self) -> dict[str, ApiSecretChangeNestedResource]:
+        objects = {}
+        for f in self._nested_objects():
+            obj = getattr(self, f)
+            if isinstance(obj, ApiSecretChangeNestedResource):
+                objects[f] = obj
+
+        return objects
+
 
 class ResourceGraph:
     """
@@ -391,8 +580,8 @@ class ResourceGraph:
 
     A node in the graph represents an instance (identified by the resource name)
     of some resource like Window, Validator etc. and connected to that node are
-    any child nodes - e.g. a Credential node will have child nodes of type Source
-    and Destination, while a Source node will have child nodes of type Window,
+    any child nodes - e.g. a Credential node will have child nodes of type
+    Source, while a Source node will have child nodes of type Window,
     Segmentation, Validator.
 
     The Resource graph is the canonical datastructure that we work with - both
@@ -424,24 +613,35 @@ class ResourceGraph:
         # Resource names are unique per resource type. This is used to track duplicate
         # names assignments to resources.
         self._resource_names_by_type: dict[str, set[str]] = {}
+        self._deprecations: list[ResourceDeprecation] = []
+
+    def add_deprecation(self, deprecation: ResourceDeprecation) -> None:
+        """Add a deprecation warning for a resource.
+
+        :param deprecation: Deprecation description
+        """
+        if self._deprecations is None:
+            self._deprecations = []
+
+        self._deprecations.append(deprecation)
 
     # Remembers that this name is in use for the specified resource type.
     # Throws an error if the specified name has been previously used.
     # the resource type.
-    def allocate_resource_name(self, resource_name: str, resource: "Resource"):
+    def allocate_resource_name(self, resource_name: str, resource: "Resource") -> None:
         resource_type: str = resource.resource_class_name()
         if resource_type not in self._resource_names_by_type:
             self._resource_names_by_type[resource_type] = set({})
 
         if resource_name in self._resource_names_by_type[resource_type]:
-            raise RuntimeError(
+            raise ValidioError(
                 f"duplicate name '{resource_name}' for type '{resource_type}'; resource"
                 " names should be unique for the same resource type"
             )
 
         self._resource_names_by_type[resource_type].add(resource_name)
 
-    def _add_root(self, resource: "Resource"):
+    def _add_root(self, resource: "Resource") -> None:
         """Adds a node to the root of a subgraph. Essentially creating
         a new subgraph.
         """
@@ -474,29 +674,30 @@ class ResourceGraph:
     def _decode(obj: dict[str, dict[str, Any]]) -> tuple["ResourceGraph", DiffContext]:
         from validio_sdk.resource.channels import Channel
         from validio_sdk.resource.credentials import Credential
-        from validio_sdk.resource.identity_providers import IdentityProvider
-        from validio_sdk.resource.users import User
+        from validio_sdk.resource.tags import Tag
         from validio_sdk.resource.validators import Validator
 
         # We decode a graph in two passes.
         # The first pass through the input does not decode
         # validators - this bit is where we transition into a non-linear
         # DAG structure because a validator can have several parents (it's
-        # target, reference source + destination). So we need to resolve all potential
-        # sources before we can resolve validators in the second pass.
+        # target, reference source). So we need to resolve all potential sources
+        # before we can resolve validators in the second pass.
         g = ResourceGraph()
 
         ctx = DiffContext()
 
         # Pass 1 - decode credentials, channels, etc. which are at the 'root'
         # level and descends onto child resources.
+        # Tags are decoded first to allow attaching them on any object.
+        ResourceGraph._decode_root(obj, ctx, g, Tag, "tags")
+
         ResourceGraph._decode_root(obj, ctx, g, Credential, "credentials")
+
         # We decode the channels graph after the credentials one - since notification
         # rules depend on sources - so that the sources are resolved by the time decode
         # the rules.
         ResourceGraph._decode_root(obj, ctx, g, Channel, "channels")
-        ResourceGraph._decode_root(obj, ctx, g, IdentityProvider, "identity_providers")
-        ResourceGraph._decode_root(obj, ctx, g, User, "users")
 
         # Pass 2 Decode validators.
         Validator._decode_pending(ctx)
@@ -510,7 +711,7 @@ class ResourceGraph:
         g: "ResourceGraph",
         resource_cls: type,
         resource_module_name: str,
-    ):
+    ) -> None:
         if resource_cls.__name__ not in obj["sub_graphs"]:
             return
 
@@ -522,7 +723,7 @@ class ResourceGraph:
             cls = eval(
                 f"validio_sdk.resource.{resource_module_name}.{v[NODE_TYPE_FIELD_NAME]}"
             )
-            resource = cast(Any, resource_cls)._decode(ctx, cls, v, g)
+            resource = cast(Any, cls)._decode(ctx, cls, v, g)
 
             resources = getattr(ctx, resource_module_name)
             resources[resource.name] = resource
