@@ -16,7 +16,7 @@ Upon exit of the program, we serialize this instance to the parent process.
 RESOURCE_GRAPH: ResourceGraph = ResourceGraph()
 
 
-def _dump_graph():
+def _dump_graph() -> None:
     # Since we piggyback on stdout, we prefix the graph with a
     # preamble to identify the start of the relevant info in the stream.
     if dump_graph_var in os.environ:
