@@ -1,6 +1,9 @@
 from dataclasses import dataclass
+from typing import Any
 
-from validio_sdk.graphql_client import GraphQLClientHttpError
+from gql.transport.exceptions import TransportQueryError, TransportServerError
+
+from validio_sdk.exception import ValidioError
 
 
 @dataclass
@@ -16,13 +19,26 @@ class SourceSchemaReinference:
 
 
 def _sanitize_error(
-    e: GraphQLClientHttpError,
+    e: TransportQueryError | TransportServerError,
     show_secrets: bool,
 ) -> Exception:
     if show_secrets:
         return e
-    raise RuntimeError(
-        f"API error: status code {e.status_code}: The error message has been "
+
+    code = ""
+    if isinstance(e, TransportServerError):
+        code = f" ({e.code})"
+
+    raise ValidioError(
+        f"API error{code}: The error message has been "
         "suppressed because it potentially contains sensitive information; "
         "If you would like to view the error message, run again with --show-secrets"
     )
+
+
+def _rename_dict_key(d: dict[str, Any], from_key: str, to_key: str) -> None:
+    if from_key not in d:
+        return
+
+    d[to_key] = d[from_key]
+    del d[from_key]
